/* BUGREPIF.C: FMH BugRep interface routines
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdarg.h>
#include "bugrepif.h"

#define MAILBOX     "\\\\.\\mailslot\\fmh-bugrep-v1"
//#define MAILBOX "c:\\mailbox.txt"
#define ALLOC_STEP  0x1000

#define MODE_BEFORE_TEXT  0 /* before anything has been added in MsgBox */
#define MODE_BOX_TEXT     1 /* text has been submitted for text field */
#define MODE_BOX_LIST     2 /* text has been submitted for list box */
#define MODE_EMAIL        3 /* text for email has been submitted */
#define MODE_FILE         4 /* file/attachment OK now */
#define MODE_ATTACH       5 /* attachment currently open */

static HANDLE curMem;             /* current memory handle */
static char *curRep;              /* current report buffer */
static int curRepSize,curRepMax;  /* report size/size of buffer */
static int curMode;               /* current curMode report is in */
static int sizeIndex;             /* index of attachment size in curRep */

int brActive(void)
/* checks wether FMH BugRep is running and active or not
   returns !=0 if active, ==0 if not
*/
 {
  HANDLE hFile;

  /* try to open mailbox */
  hFile=CreateFile(MAILBOX,GENERIC_WRITE,FILE_SHARE_READ,
                   NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL); 
  
  /* return result code if unavailable */
  if (hFile==INVALID_HANDLE_VALUE) return 0;

  /* close mailbox without sending anything */
  CloseHandle(hFile);

  /* FMH BugRep alive and kicking! */
  return 1;
 }

static void dByte(int val)
/* writes a single byte to current report */
 {
  /* allocate more memory if necessary */
  if (curRepSize==curRepMax) {
    if (!curRep) {
      curMem=GlobalAlloc(GMEM_FIXED,ALLOC_STEP);
      curRep=GlobalLock(curMem);
      }
    else {
      GlobalUnlock(curMem);
      curMem=GlobalReAlloc(curMem,curRepMax+ALLOC_STEP,0);
      curRep=GlobalLock(curMem);
      }
    if (!curRep) return;
    curRepMax+=ALLOC_STEP;
    }

  /* now write byte */
  curRep[curRepSize++]=(char)val;
 }

static void dString(char *s)
/* writes string to current report (without \0) */
 {
  while (*s) dByte(*s++);
 }

void brNew(void)
/* starts a new bug report */
 {
  curRep=NULL;
  curRepSize=curRepMax=0;
  curMode=MODE_BEFORE_TEXT;

  /* init header */
  dByte(1);
 }

#define GET_TEXT()  \
  va_list va; \
  char buf[500]; \
  \
  va_start(va,text); \
  vsprintf(buf,text,va); \
  va_end(va)

void brMsgBoxText(char *text, ...)
/* adds text to the error message box (normal text field) */
 {
  GET_TEXT();

  /* curMode OK? */
  if (!*buf||curMode>MODE_BOX_TEXT) return;

  /* correct curMode, write text */
  curMode=MODE_BOX_TEXT;
  dString(buf);
 }

void brMsgBoxList(char *text, ...)
/* like before, but for the list box; submits one line per call */
 {
  GET_TEXT();

  /* curMode OK? */
  if (curMode>MODE_BOX_LIST) return;

  /* append \0 to box text if necessary */
  switch (curMode) {
    case MODE_BEFORE_TEXT:
      dByte(' ');
    case MODE_BOX_TEXT:
      dByte(0);
      curMode=MODE_BOX_LIST;
      break;
    }

  /* write text */
  dString(*buf?buf:" ");
  dByte(0);
 }

void brMailText(char *text, ...)
/* submits text for email message (ISO charset) */
 {
  GET_TEXT();

  /* curMode OK? */
  if (!*buf||curMode>MODE_EMAIL) return;

  /* get to current curMode */
  switch (curMode) {
    case MODE_BOX_TEXT:
      dByte(0);
    case MODE_BOX_LIST:
    case MODE_BEFORE_TEXT:
      dByte(0); 
      curMode=MODE_EMAIL;
      break;
    }

  /* write text */
  dString(buf);
 }

void brMailFile(char *attFn, char *realFn)
/* attaches an external file */
 {
  /* get to current curMode */
  switch (curMode) {
    case MODE_BOX_TEXT:
      dByte(0);
    case MODE_BEFORE_TEXT:
    case MODE_BOX_LIST:
      dByte(0);
    case MODE_EMAIL:
      dByte(0);
      break;
    }
  curMode=MODE_FILE;

  /* write info to report */
  dString(attFn); dByte(0);
  dByte(0xff); dByte(0xff); dByte(0xff); dByte(0xff);
  dString(realFn); dByte(0);
 }

void brMailAttach(char *attFn)
/* starts attaching an internal file (program generated) */
 {
  /* get to current curMode */
  switch (curMode) {
    case MODE_BOX_TEXT:
      dByte(0);
    case MODE_BEFORE_TEXT:
    case MODE_BOX_LIST:
      dByte(0);
    case MODE_EMAIL:
      dByte(0);
      break;
    }
  curMode=MODE_ATTACH;

  /* write info to report */
  dString(attFn); dByte(0);
  sizeIndex=curRepSize;
  dByte(0); dByte(0); dByte(0); dByte(0);
 }

void brMailAttachData(void *data, int size)
/* adds data to the current internal file */
 {
  char *ptr;

  /* curMode OK? */
  if (curMode!=MODE_ATTACH) return;

  /* correct size */
  *(unsigned *)(curRep+sizeIndex)+=size;

  /* write data */
  ptr=(char *)data;
  while (size-->0) dByte(*ptr++);
 }

void brSubmit(void)
/* submits current bug report */
 {
  HANDLE hFile;

  if (!curRep) return;

  /* finish current report */
  switch (curMode) {
    case MODE_BOX_TEXT:
      dByte(0);
    case MODE_BEFORE_TEXT:
    case MODE_BOX_LIST:
      dByte(0);
    case MODE_EMAIL:
      dByte(0);
      break;
    }
  dByte(0);

  /* open mailbox */
  hFile=CreateFile(MAILBOX,GENERIC_WRITE,FILE_SHARE_READ,
                   NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL); 
  if (hFile!=INVALID_HANDLE_VALUE) {
    DWORD wr;

    /* submit report */
    WriteFile(hFile,curRep,curRepSize,&wr,NULL);

    /* close mailbox */
    CloseHandle(hFile);
    }

  /* free allocated memory */
  GlobalUnlock(curMem);
  GlobalFree(curMem);
 }
