// EXHNDLER.C: Exception handler

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "exhndler.h"
#include "cvload.h"
#include "bugrepif.h"

/*
    Information given by this module:
    ---------------------------------

    - exception code and string
    - adress where exception occured
    - register values
    - stack walk, showing: address, function, source file and line
*/

// file-global static data
static LPTOP_LEVEL_EXCEPTION_FILTER prevFilter;
static int useCV;

static char *getExceptionString(DWORD dwCode)
// Given an exception code, returns a pointer to a static string with a 
// description of the exception                                         
{
  static char buf[512];
  #define EXCEPTION( x ) case EXCEPTION_##x: return #x;

  switch ( dwCode )
  {
    EXCEPTION( ACCESS_VIOLATION )
    EXCEPTION( DATATYPE_MISALIGNMENT )
    EXCEPTION( BREAKPOINT )
    EXCEPTION( SINGLE_STEP )
    EXCEPTION( ARRAY_BOUNDS_EXCEEDED )
    EXCEPTION( FLT_DENORMAL_OPERAND )
    EXCEPTION( FLT_DIVIDE_BY_ZERO )
    EXCEPTION( FLT_INEXACT_RESULT )
    EXCEPTION( FLT_INVALID_OPERATION )
    EXCEPTION( FLT_OVERFLOW )
    EXCEPTION( FLT_STACK_CHECK )
    EXCEPTION( FLT_UNDERFLOW )
    EXCEPTION( INT_DIVIDE_BY_ZERO )
    EXCEPTION( INT_OVERFLOW )
    EXCEPTION( PRIV_INSTRUCTION )
    EXCEPTION( IN_PAGE_ERROR )
    EXCEPTION( ILLEGAL_INSTRUCTION )
    EXCEPTION( NONCONTINUABLE_EXCEPTION )
    EXCEPTION( STACK_OVERFLOW )
    EXCEPTION( INVALID_DISPOSITION )
    EXCEPTION( GUARD_PAGE )
    EXCEPTION( INVALID_HANDLE )
  }

  // If not one of the "known" exceptions, try to get the string
  // from NTDLL.DLL's message table.
  *buf=0;
  FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_FROM_HMODULE,
                GetModuleHandle("NTDLL.DLL"),
                dwCode,0,buf,sizeof(buf),0);
  return buf;
}

static BOOL getLogicalAddress(PVOID addr, PTSTR szModule, DWORD len, 
                       DWORD *section, DWORD *offset)
// Given a linear address, locates the module, section, and offset containing  
// that address. Code by Matt Pietrek.
{
  MEMORY_BASIC_INFORMATION mbi;
  DWORD hMod;
  PIMAGE_DOS_HEADER pDosHdr;
  PIMAGE_NT_HEADERS pNtHdr;
  PIMAGE_SECTION_HEADER pSection;
  DWORD rva;
  unsigned i;

  if (!VirtualQuery(addr,&mbi,sizeof(mbi)))
    return FALSE;

  hMod=(DWORD)mbi.AllocationBase;
  if (!GetModuleFileName((HMODULE)hMod,szModule,len))
    return FALSE;

  // Point to the DOS header in memory
  pDosHdr=(PIMAGE_DOS_HEADER)hMod;

  // From the DOS header, find the NT (PE) header
  pNtHdr=(PIMAGE_NT_HEADERS)(hMod+pDosHdr->e_lfanew);
  pSection=IMAGE_FIRST_SECTION(pNtHdr);
  rva=(DWORD)addr-hMod; // RVA is offset from module load address

  // Iterate through the section table, looking for the one that encompasses
  // the linear address.
  for (i=0;i<pNtHdr->FileHeader.NumberOfSections;i++,pSection++)
  {
    DWORD sectionStart=pSection->VirtualAddress;
    DWORD sectionEnd=sectionStart+max(pSection->SizeOfRawData,pSection->Misc.VirtualSize);

    // Is the address in this section???
    if ((rva>=sectionStart)&&(rva<=sectionEnd))
    {
      // Yes, address is in the section.  Calculate section and offset,
      // and store in the "section" & "offset" params, which were
      // passed by reference.
      *section=i+1;
      *offset=rva-sectionStart;
      return TRUE;
    }
  }

  return FALSE; // Should never get here!
}

static void genMsgBox(PEXCEPTION_POINTERS pe)
{
  char faultMod[MAX_PATH];
  DWORD sec,ofs;
  PCONTEXT pCtx=pe->ContextRecord;
  PEXCEPTION_RECORD pEr=pe->ExceptionRecord; 

  // some blabla for the user
  brMsgBoxText("The current process has generated a general protection fault "
               "and has been terminated. You may send a bug report via email "
               "if you want to.");

  // a little specific information goes in here:
  brMsgBoxList("Exception code: 0x%08x %s",pEr->ExceptionCode,
                  getExceptionString(pEr->ExceptionCode));

  // where has the fault occured?
  if (getLogicalAddress(pEr->ExceptionAddress,faultMod,sizeof(faultMod),&sec,&ofs))
    brMsgBoxList("Fault address:  0x%08x 0x%02x:0x%08x %s",
                    pEr->ExceptionAddress,sec,ofs,faultMod);
  else
    brMsgBoxList("Fault address:  0x%08x",pEr->ExceptionAddress);

  // here comes the register dump!
  brMsgBoxList("");
  brMsgBoxList("Registers:");

  brMsgBoxList("EAX: 0x%08x\tCS: 0x%04x\tEIP: 0x%08x\tFLG: 0x%08x",
                  pCtx->Eax,pCtx->SegCs,pCtx->Eip,pCtx->EFlags);
  brMsgBoxList("EBX: 0x%08x\tSS: 0x%04x\tESP: 0x%08x\tEBP: 0x%08x",
                  pCtx->Ebx,pCtx->SegSs,pCtx->Esp,pCtx->Ebp);
  brMsgBoxList("ECX: 0x%08x\tDS: 0x%04x\tESI: 0x%08x\tFS: 0x%04x",
                  pCtx->Ecx,pCtx->SegDs,pCtx->Esi,pCtx->SegFs);
  brMsgBoxList("EDX: 0x%08x\tES: 0x%04x\tEDI: 0x%08x\tGS: 0x%04x",
                  pCtx->Edx,pCtx->SegEs,pCtx->Edi,pCtx->SegGs);
}

static void genBugReport(PEXCEPTION_POINTERS pe)
{
  char faultMod[MAX_PATH];
  DWORD sec,ofs,pc;
  PCONTEXT pCtx=pe->ContextRecord;
  PDWORD pFrame,pPrevFrame;
  PEXCEPTION_RECORD pEr=pe->ExceptionRecord; 

  // send same like above...
  brMailText("Exception code: 0x%08x %s\n",pEr->ExceptionCode,
                  getExceptionString(pEr->ExceptionCode));

  // where has the fault occured?
  if (getLogicalAddress(pEr->ExceptionAddress,faultMod,sizeof(faultMod),&sec,&ofs))
    brMailText("Fault address:  0x%08x 0x%02x:0x%08x %s\n",
                    pEr->ExceptionAddress,sec,ofs,faultMod);
  else
    brMailText("Fault address:  0x%08x\n",pEr->ExceptionAddress);

  // here comes the register dump!
  brMailText("\nRegisters:\n");

  brMailText("EAX: 0x%08x   CS: 0x%04x   EIP: 0x%08x   FLG: 0x%08x\n",
                  pCtx->Eax,pCtx->SegCs,pCtx->Eip,pCtx->EFlags);
  brMailText("EBX: 0x%08x   SS: 0x%04x   ESP: 0x%08x   EBP: 0x%08x\n",
                  pCtx->Ebx,pCtx->SegSs,pCtx->Esp,pCtx->Ebp);
  brMailText("ECX: 0x%08x   DS: 0x%04x   ESI: 0x%08x   FS: 0x%04x\n",
                  pCtx->Ecx,pCtx->SegDs,pCtx->Esi,pCtx->SegFs);
  brMailText("EDX: 0x%08x   ES: 0x%04x   EDI: 0x%08x   GS: 0x%04x\n\n",
                  pCtx->Edx,pCtx->SegEs,pCtx->Edi,pCtx->SegGs);

  // now walk the stack...
  brMailText("Call stack:\n");
  brMailText(" Address   Frame   Logical addr Module\n");

  pc=pCtx->Eip;
  pFrame=(PDWORD)pCtx->Ebp;

  do
  {
    char mod[MAX_PATH]="";
    DWORD sec=0,ofs=0;
    int dumpStack=0;

    if (getLogicalAddress((PVOID)pc,mod,sizeof(mod),&sec,&ofs))
      brMailText("%08x %08x %04x:%08x %s\n",pc,pFrame,sec,ofs,mod);
    else
      brMailText("%08x %08x\n",pc,pFrame);

    if (useCV&&sec==1) 
    {
      // get extra info using codeview debug information
      tCVQuery q;

      if (cvQuery(ofs,&q)) 
      {
        dumpStack=1;
        brMailText("%18sProc: %s+0x%04x (0x%08x to 0x%08x)\n",
                      "",q.procName,ofs-q.procFrom,q.procFrom,q.procTo);
        if (q.srcName)
        {
          brMailText("%18sSource: %s+0x%04x\n%24s(0x%08x to 0x%08x)\n",
                      "",q.srcName,ofs-q.srcFrom,"",q.srcFrom,q.srcTo);
          if (q.line)
            brMailText("%18sLine: %i (0x%08x)\n",
                      "",q.line,q.lineOfs);
        }
      }
    }

    pc=pFrame[1];
    pPrevFrame=pFrame;

    pFrame=(PDWORD)pFrame[0]; // precede to next higher frame on stack

    if ((DWORD)pFrame&3)  // Frame pointer must be aligned on a
      break;              // DWORD boundary.  Bail if not so.

    if (pFrame<=pPrevFrame) break;

    // Can two DWORDs be read from the supposed frame address?          
    if (IsBadWritePtr(pFrame, sizeof(PVOID)*2)) break;

    if (dumpStack) 
    {
      DWORD *help=pPrevFrame;
      int k=0;

      while (help<pFrame) 
      {
        if (!k) 
        {
          brMailText("%10sEBP+0x%04x: ","",help-pPrevFrame);
          if (help==pPrevFrame)
          {
            help+=2;
            brMailText("%18s","");
            k+=2;
          }
        }
        if (IsBadWritePtr(help,sizeof(PVOID)))
          brMailText("******** ");
        else brMailText("%08x ",*help);
        help++;
        if (++k==4) 
        {
          brMailText("\n");
          k=0;
        }
      }
      if (k) brMailText("\n");
    }

    } while (1);
}

static LONG WINAPI customFilter(PEXCEPTION_POINTERS pe)
// This is the custom top level exception handler. If FMH BugRep is
// not installed or not running, it will simply pass the exception
// to the old handler. Otherwise it will gather some information about
// the running process and then exit the process.
{
  // is FMH BugRep running?
  if (brActive()) 
  {
    UINT oldMode;

    // turn off windows GPF error box (BugRep will show error box instead)
    oldMode=SetErrorMode(0);
    SetErrorMode(oldMode|SEM_NOGPFAULTERRORBOX);

    // start bug report
    brNew();

    // generate text for message box
    genMsgBox(pe);

    // generate text for bug report
    genBugReport(pe);

    // submit report
    brSubmit();
  }

  // let Windows do the rest
  if (prevFilter)
    return prevFilter(pe);
  else
    return EXCEPTION_CONTINUE_SEARCH;
}

void exHandlerInstall(void)
// install custom top level exception handler
{
  // load codeview info
  useCV=cvLoad(NULL);

  prevFilter=SetUnhandledExceptionFilter(customFilter);
}
